#MIT License

#Copyright (c) 2023 Adam Hines, Peter G Stratton, Michael Milford, Tobias Fischer

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'''
Imports
'''
import os
import sys
import torch
import argparse

import torch.quantization as quantization

from tqdm import tqdm
from vprtempo.VPRTempo import VPRTempo, run_inference
from vprtempo.VPRTempoTrain import VPRTempoTrain, train_new_model
from vprtempo.src.loggers import model_logger, model_logger_quant
from vprtempo.VPRTempoQuant import VPRTempoQuant, run_inference_quant
from vprtempo.VPRTempoQuantTrain import VPRTempoQuantTrain, train_new_model_quant

def generate_model_name(model,quant=False):
    """
    Generate the model name based on its parameters.
    """
    if quant:
        model_name = (''.join(model.database_dirs)+"_"+
                "VPRTempoQuant_" +
                "IN"+str(model.input)+"_" +
                "FN"+str(model.feature)+"_" + 
                "DB"+str(model.database_places) +
                ".pth")
    else:
        model_name = (''.join(model.database_dirs)+"_"+
                "VPRTempo_" +
                "IN"+str(model.input)+"_" +
                "FN"+str(model.feature)+"_" + 
                "DB"+str(model.database_places) +
                ".pth")
    return model_name

def check_pretrained_model(model_name):
    """
    Check if a pre-trained model exists and prompt the user to retrain if desired.
    """
    if os.path.exists(os.path.join('./vprtempo/models', model_name)):
        prompt = "A network with these parameters exists, re-train network? (y/n):\n"
        retrain = input(prompt).strip().lower()
        if retrain == 'y':
            return True
        elif retrain == 'n':
            print('Training new model cancelled')
            sys.exit()

def initialize_and_run_model(args,dims):
    """
    Run the VPRTempo/VPRTempoQuant training or inference models.
    
    :param args: Arguments set for the network
    :param dims: Dimensions of the network
    """
    # Determine number of modules to generate based on user input
    places = args.database_places # Copy out number of database places

    # Caclulate number of modules
    num_modules = 1
    while places > args.max_module:
        places -= args.max_module
        num_modules += 1

    # If the final module has less than max_module, reduce the dim of the output layer
    remainder = args.database_places % args.max_module
    if remainder != 0: # There are remainders, adjust output neuron count in final module
        out_dim = int((args.database_places - remainder) / (num_modules - 1))
        final_out_dim = remainder
    else: # No remainders, all modules are even
        out_dim = int(args.database_places / num_modules)
        final_out_dim = out_dim

    # If user wants to train a new network
    if args.train_new_model:
        # If using quantization aware training
        if args.quantize:
            models = []
            logger = model_logger_quant() # Initialize the logger
            qconfig = quantization.get_default_qat_qconfig('fbgemm')
            # Create the modules    
            final_out = None
            for mod in tqdm(range(num_modules), desc="Initializing modules"):
                model = VPRTempoQuantTrain(args, dims, logger, num_modules, out_dim, out_dim_remainder=final_out) # Initialize the model
                model.train()
                model.qconfig = qconfig
                quantization.prepare_qat(model, inplace=True)
                models.append(model) # Create module list
                if mod == num_modules - 2:
                    final_out = final_out_dim
            # Generate the model name
            model_name = generate_model_name(model,args.quantize)
            # Check if the model has been trained before
            check_pretrained_model(model_name)
            # Get the quantization config
            qconfig = quantization.get_default_qat_qconfig('fbgemm')
            # Train the model
            train_new_model_quant(models, model_name)

        # Base model    
        else:
            models = []
            logger = model_logger() # Initialize the logger

            # Create the modules    
            final_out = None
            for mod in tqdm(range(num_modules), desc="Initializing modules"):
                model = VPRTempoTrain(args, dims, logger, num_modules, out_dim, out_dim_remainder=final_out) # Initialize the model
                model.to(torch.device('cpu')) # Move module to CPU for storage (necessary for large models)
                models.append(model) # Create module list
                if mod == num_modules - 2:
                    final_out = final_out_dim

            # Generate the model name
            model_name = generate_model_name(model)
            print(f"Model name: {model_name}")
            # Check if the model has been trained before
            check_pretrained_model(model_name)
            # Train the model
            train_new_model(models, model_name)

    # Run the inference network
    else:
        # Set the quantization configuration
        if args.quantize:
            models = []
            logger, output_folder = model_logger_quant()
            qconfig = quantization.get_default_qat_qconfig('fbgemm')
            final_out = None
            for _ in tqdm(range(num_modules), desc="Initializing modules"):
                # Initialize the model
                model = VPRTempoQuant(
                    args,
                    dims,
                    logger,
                    num_modules,
                    output_folder,
                    out_dim,
                    out_dim_remainder=final_out
                    ) 
                model.eval()
                model.qconfig = qconfig
                quantization.prepare(model, inplace=True)
                quantization.convert(model, inplace=True)
                models.append(model)
            # Generate the model name
            model_name = generate_model_name(model, args.quantize)
            # Run the quantized inference model
            run_inference_quant(models, model_name)
        else:
            models = []
            logger, output_folder = model_logger() # Initialize the logger
            places = args.database_places # Copy out number of database places

            # Create the modules    
            final_out = None
            for mod in tqdm(range(num_modules), desc="Initializing modules"):
                model = VPRTempo(
                    args,
                    dims,
                    logger,
                    num_modules,
                    output_folder,
                    out_dim,
                    out_dim_remainder=final_out
                    ) 
                model.eval()
                model.to(torch.device('cpu')) # Move module to CPU for storage (necessary for large models)
                models.append(model) # Create module list
                if mod == num_modules - 2:
                    final_out = final_out_dim
            # Generate the model name
            model_name = generate_model_name(model)
            print(f"Model name: {model_name}")
            # Run the inference model
            run_inference(models, model_name)

def parse_network(use_quantize=False, train_new_model=False):
    '''
    Define the base parameter parser (configurable by the user)
    '''
    parser = argparse.ArgumentParser(description="Args for base configuration file")

    # Define the dataset arguments
    parser.add_argument('--dataset', type=str, default='nordland',
                            help="Dataset to use for training and/or inferencing")
    parser.add_argument('--data_dir', type=str, default='./vprtempo/dataset/',
                            help="Directory where dataset files are stored")
    parser.add_argument('--database_places', type=int, default=500,
                            help="Number of places to use for training")
    parser.add_argument('--query_places', type=int, default=500,
                            help="Number of places to use for inferencing")
    parser.add_argument('--max_module', type=int, default=500,
                            help="Maximum number of images per module")
    parser.add_argument('--database_dirs', type=str, default='spring,fall',
                            help="Directories to use for training")
    parser.add_argument('--query_dir', type=str, default='summer',
                            help="Directories to use for testing")
    parser.add_argument('--GT_tolerance', type=int, default=1,
                            help="Ground truth tolerance for matching")
    parser.add_argument('--skip', type=int, default=4799,
                            help="Images to skip for training and/or inferencing")

    # Define training parameters
    parser.add_argument('--filter', type=int, default=8,
                            help="Images to skip for training and/or inferencing")
    parser.add_argument('--epoch', type=int, default=4,
                            help="Number of epochs to train the model")

    # Define image transformation parameters
    parser.add_argument('--patches', type=int, default=15,
                            help="Number of patches to generate for patch normalization image into")
    parser.add_argument('--dims', type=str, default="56,56",
                        help="Dimensions to resize the image to")

    # Define the network functionality
    parser.add_argument('--train_new_model', action='store_true',
                            help="Flag to run the training or inferencing model")
    parser.add_argument('--quantize', action='store_true',
                            help="Enable/disable quantization for the model")
    
    # Define metrics functionality
    parser.add_argument('--PR_curve', action='store_true',
                            help="Flag to generate a Precision-Recall curve")
    parser.add_argument('--sim_mat', action='store_true',
                            help="Flag to plot the similarity matrix, GT, and GTsoft")
    
    # If the function is called with specific arguments, override sys.argv
    if use_quantize or train_new_model:
        sys.argv = ['']
        if use_quantize:
            sys.argv.append('--quantize')
        if train_new_model:
            sys.argv.append('--train_new_model')

    # Output base configuration
    args = parser.parse_args()
    dims = [int(x) for x in args.dims.split(",")]

    # Run the network with the desired settings
    initialize_and_run_model(args,dims)

if __name__ == "__main__":
    # User input to determine if using quantized network or to train new model 
    parse_network(use_quantize=False, 
                  train_new_model=False)


import os
from PIL import Image
from tqdm import tqdm
import image
from camera_model import CameraModel

# Set this to the directory containing the images
base_path = '/media/adam/vprdatasets/data/orc'
# Modify if evaluating different ORC datasets
datasets = [
            '2015-08-12-15-04-18',
            '2014-11-21-16-07-03',
            '2015-10-29-12-18-17'
        ]
# Set up output path names
processed_path = 'demosaiced'

for dataset in datasets:
    # Define the dataset path
    dataset_path = os.path.join(base_path, dataset)

    # Create the folders
    os.makedirs(os.path.join(dataset_path, processed_path), exist_ok=True)

    # file path for the robotcar-dataset-sdk models folder
    model_dir = '/home/adam/repo/robotcar-dataset-sdk/models/'
    # file path for the left stereo images
    images_path = os.path.join(dataset_path, 'stereo/left')
    # Create a camera model object
    model = CameraModel(model_dir,images_path)
    # Get sorted list of PNG images
    images = sorted([os.path.join(images_path, image) for image in os.listdir(images_path) if image.endswith('.png')])

    # Process each image with a progress bar
    for img in tqdm(images, desc="Processing images", unit="image"):
        # Load the image
        processed_img = image.load_image(img, model=model)
        # Create the output file path
        output_img = os.path.join(dataset_path, processed_path, os.path.basename(img))
        processed_img = Image.fromarray(processed_img, mode="RGB")
        # Save the processed image as PNG
        processed_img.save(output_img, "PNG")

    print("Images processed and saved successfully.")


#   =====================================================================
#   Copyright (C) 2023  Stefan Schubert, stefan.schubert@etit.tu-chemnitz.de
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#   =====================================================================
#
import numpy as np


def createPR(S_in, GThard, GTsoft=None, matching='multi', n_thresh=100):
    """
    Calculates the precision and recall at n_thresh equally spaced threshold values
    for a given similarity matrix S_in and ground truth matrices GThard and GTsoft for
    single-best-match VPR or multi-match VPR.

    The matrices S_in, GThard and GTsoft are two-dimensional and should all have the
    same shape.
    The matrices GThard and GTsoft should be binary matrices, where the entries are
    only zeros or ones.
    The matrix S_in should have continuous values between -Inf and Inf. Higher values
    indicate higher similarity.
    The string matching should be set to either "single" or "multi" for single-best-
    match VPR or multi-match VPR.
    The integer n_tresh controls the number of threshold values and should be >1.
    """

    assert (S_in.shape == GThard.shape and S_in.shape == GTsoft.shape),"S_in, GThard and GTsoft must have the same shape"
    assert (S_in.ndim == 2),"S_in, GThard and GTsoft must be two-dimensional"
    assert (matching in ['single', 'multi']),"matching should contain one of the following strings: [single, multi]"
    assert (n_thresh > 1),"n_thresh must be >1"

    # ensure logical datatype in GT and GTsoft
    GT = GThard.astype('bool')
    if GTsoft is not None:
        GTsoft = GTsoft.astype('bool')

    # copy S and set elements that are only true in GTsoft to min(S) to ignore them during evaluation
    S = S_in.copy()
    if GTsoft is not None:
        S[GTsoft & ~GT] = S.min()

    # single-best-match or multi-match VPR
    if matching == 'single':
        # count the number of ground-truth positives (GTP)
        GTP = np.count_nonzero(GT.any(0))

        # GT-values for best match per query (i.e., per column)
        GT = GT[np.argmax(S, axis=0), np.arange(GT.shape[1])]

        # similarities for best match per query (i.e., per column)
        S = np.max(S, axis=0)

    elif matching == 'multi':
        # count the number of ground-truth positives (GTP)
        GTP = np.count_nonzero(GT) # ground truth positives

    # init precision and recall vectors
    R = [0, ]
    P = [1, ]

    # select start and end treshold
    startV = S.max()  # start-value for treshold
    endV = S.min()  # end-value for treshold

    # iterate over different thresholds
    for i in np.linspace(startV, endV, n_thresh):
        B = S >= i  # apply threshold

        TP = np.count_nonzero(GT & B)  # true positives
        FP = np.count_nonzero((~GT) & B)  # false positives

        P.append(TP / (TP + FP))  # precision
        R.append(TP / GTP)  # recall

    return P, R


def recallAt100precision(S_in, GThard, GTsoft=None, matching='multi', n_thresh=100):
    """
    Calculates the maximum recall at 100% precision for a given similarity matrix S_in 
    and ground truth matrices GThard and GTsoft for single-best-match VPR or multi-match 
    VPR.

    The matrices S_in, GThard and GTsoft are two-dimensional and should all have the
    same shape.
    The matrices GThard and GTsoft should be binary matrices, where the entries are
    only zeros or ones.
    The matrix S_in should have continuous values between -Inf and Inf. Higher values
    indicate higher similarity.
    The string matching should be set to either "single" or "multi" for single-best-
    match VPR or multi-match VPR.
    The integer n_tresh controls the number of threshold values during the creation of
    the precision-recall curve and should be >1.
    """

    assert (S_in.shape == GThard.shape),"S_in and GThard must have the same shape"
    if GTsoft is not None:
        assert (S_in.shape == GTsoft.shape),"S_in and GTsoft must have the same shape"
    assert (S_in.ndim == 2),"S_in, GThard and GTsoft must be two-dimensional"
    assert (matching in ['single', 'multi']),"matching should contain one of the following strings: [single, multi]"
    assert (n_thresh > 1),"n_thresh must be >1"

    # get precision-recall curve
    P, R = createPR(S_in, GThard, GTsoft, matching=matching, n_thresh=n_thresh)
    P = np.array(P)
    R = np.array(R)

    # recall values at 100% precision
    R = R[P==1]

    # maximum recall at 100% precision
    R = R.max()

    return R


def recallAtK(S_in, GThard, GTsoft=None, K=1):
    """
    Calculates the recall@K for a given similarity matrix S_in and ground truth matrices 
    GThard and GTsoft.

    The matrices S_in, GThard and GTsoft are two-dimensional and should all have the
    same shape.
    The matrices GThard and GTsoft should be binary matrices, where the entries are
    only zeros or ones.
    The matrix S_in should have continuous values between -Inf and Inf. Higher values
    indicate higher similarity.
    The integer K>=1 defines the number of matching candidates that are selected and
    that must contain an actually matching image pair.
    """

    assert (S_in.shape == GThard.shape),"S_in and GThard must have the same shape"
    if GTsoft is not None:
        assert (S_in.shape == GTsoft.shape),"S_in and GTsoft must have the same shape"
    assert (S_in.ndim == 2),"S_in, GThard and GTsoft must be two-dimensional"
    assert (K >= 1),"K must be >=1"

    # ensure logical datatype in GT and GTsoft
    GT = GThard.astype('bool')
    if GTsoft is not None:
        GTsoft = GTsoft.astype('bool')

    # copy S and set elements that are only true in GTsoft to min(S) to ignore them during evaluation
    S = S_in.copy()
    if GTsoft is not None:
        S[GTsoft & ~GT] = S.min()

    # discard all query images without an actually matching database image
    j = GT.sum(0) > 0 # columns with matches
    S = S[:,j] # select columns with a match
    GT = GT[:,j] # select columns with a match

    # select K highest similarities
    i = S.argsort(0)[-K:,:]
    j = np.tile(np.arange(i.shape[1]), [K, 1])
    GT = GT[i, j]

    # recall@K
    RatK = np.sum(GT.sum(0) > 0) / GT.shape[1]

    return RatK


import os
import math
import torch

import pandas as pd
import numpy as np
import torch.nn.functional as F

from torchvision.io import read_image
from torch.utils.data import Dataset

class GetPatches2D:
    def __init__(self, patch_size, image_pad):
        self.patch_size = patch_size
        self.image_pad = image_pad
    
    def __call__(self, img):

        # Assuming image_pad is already a PyTorch tensor. If not, you can convert it:
        # image_pad = torch.tensor(image_pad).to(torch.float64)

        # Using unfold to get 2D sliding windows.
        unfolded = self.image_pad.unfold(0, self.patch_size[0], 1).unfold(1, self.patch_size[1], 1)
        # The size of unfolded will be [nrows, ncols, patch_size[0], patch_size[1]]

        # Reshaping the tensor to the desired shape
        patches = unfolded.permute(2, 3, 0, 1).contiguous().view(self.patch_size[0]*self.patch_size[1], -1)

        return patches


class PatchNormalisePad:
    def __init__(self, patches):
        self.patches = patches

    
    def nanstd(self,input_tensor, dim=None, unbiased=True):
        if dim is not None:
            valid_count = torch.sum(~torch.isnan(input_tensor), dim=dim, dtype=torch.float)
            mean = torch.nansum(input_tensor, dim=dim) / valid_count
            diff = input_tensor - mean.unsqueeze(dim)
            variance = torch.nansum(diff * diff, dim=dim) / valid_count

            # Bessel's correction for unbiased estimation
            if unbiased:
                variance = variance * (valid_count / (valid_count - 1))
        else:
            valid_count = torch.sum(~torch.isnan(input_tensor), dtype=torch.float)
            mean = torch.nansum(input_tensor) / valid_count
            diff = input_tensor - mean
            variance = torch.nansum(diff * diff) / valid_count
            
            # Bessel's correction for unbiased estimation
            if unbiased:
                variance = variance * (valid_count / (valid_count - 1))

        return torch.sqrt(variance)
   
    def __call__(self, img):
        img = torch.squeeze(img,0)
        patch_size = (self.patches, self.patches)
        patch_half_size = [int((p-1)/2) for p in patch_size ]
        
        # Compute the padding. If patch_half_size is a scalar, the same value will be used for all sides.
        if isinstance(patch_half_size, int):
            pad = (patch_half_size, patch_half_size, patch_half_size, patch_half_size)  # left, right, top, bottom
        else:
            # If patch_half_size is a tuple, then we'll assume it's in the format (height, width)
            pad = (patch_half_size[1], patch_half_size[1], patch_half_size[0], patch_half_size[0])  # left, right, top, bottom

        # Apply padding
        image_pad = F.pad(img, pad, mode='constant', value=float('nan'))

        nrows = img.shape[0] 
        ncols = img.shape[1]
        patcher = GetPatches2D(patch_size,image_pad)
        patches = patcher(img)
        mus = torch.nanmean(patches, dim=0)
        stds = self.nanstd(patches, dim=0)
        with np.errstate(divide='ignore', invalid='ignore'):
            im_norm = (img - mus.reshape(nrows, ncols)) / stds.reshape(nrows, ncols)
        
        im_norm[torch.isnan(im_norm)] = 0.0
        im_norm[im_norm < -1.0] = -1.0
        im_norm[im_norm > 1.0] = 1.0
        
        return im_norm

class SetImageAsSpikes:
    def __init__(self, intensity=255, test=True):
        self.intensity = intensity
        
        # Setup QAT FakeQuantize for the activations (your spikes)
        self.fake_quantize = torch.quantization.FakeQuantize(
            observer=torch.quantization.MovingAverageMinMaxObserver, 
            quant_min=0, 
            quant_max=255, 
            dtype=torch.quint8, 
            qscheme=torch.per_tensor_affine, 
            reduce_range=False
        )
        
    def train(self):
        self.fake_quantize.train()

    def eval(self):
        self.fake_quantize.eval()    
    
    def __call__(self, img_tensor):
        N, W, H = img_tensor.shape
        reshaped_batch = img_tensor.view(N, 1, -1)
        
        # Divide all pixel values by 255
        normalized_batch = reshaped_batch / self.intensity
        normalized_batch = torch.squeeze(normalized_batch, 0)

        # Apply FakeQuantize
        spikes = self.fake_quantize(normalized_batch)
        
        if not self.fake_quantize.training:
            scale, zero_point = self.fake_quantize.calculate_qparams()
            spikes = torch.quantize_per_tensor(spikes, float(scale), int(zero_point), dtype=torch.quint8)

        return spikes

class ProcessImage:
    def __init__(self, dims, patches):
        self.dims = dims
        self.patches = patches
        
    def __call__(self, img):
        # Convert the image to grayscale using the standard weights for RGB channels
        if img.shape[0] == 3:
            img = 0.299 * img[0] + 0.587 * img[1] + 0.114 * img[2]
         # Add a channel dimension to the resulting grayscale image
        img= img.unsqueeze(0)

        # gamma correction
        mid = 0.5
        mean = torch.mean(img.float())
        gamma = math.log(mid * 255) / math.log(mean)
        img = torch.pow(img, gamma).clip(0, 255)
        
        # resize and patch normalize        
        if len(img.shape) == 3:
            img = img.unsqueeze(0)
        img = F.interpolate(img, size=self.dims, mode='bilinear', align_corners=False)
        img = img.squeeze(0)
        patch_normaliser = PatchNormalisePad(self.patches)
        im_norm = patch_normaliser(img) 
        img = (255.0 * (1 + im_norm) / 2.0).to(dtype=torch.uint8)
        img = torch.unsqueeze(img,0)
        spike_maker = SetImageAsSpikes()
        img = spike_maker(img)
        img = torch.squeeze(img,0)

        return img

class CustomImageDataset(Dataset):
    def __init__(self, annotations_file, base_dir, img_dirs, transform=None, target_transform=None, 
                 filter=1, skip=0, max_samples=None, test=True, img_range=None):
        self.transform = transform
        self.target_transform = target_transform
        self.filter = filter
        self.img_range = img_range
        self.skip = skip
        
        # Load image labels from each directory, apply the skip and max_samples, and concatenate
        self.img_labels = []
        # if annotations_file is a single file, convert it to a list
        if not isinstance(annotations_file, list):
            annotations_file = [annotations_file]
        for idx, annotation in enumerate(annotations_file):
            img_labels = pd.read_csv(annotation)
            img_labels['file_path'] = img_labels.apply(lambda row: os.path.join(base_dir,img_dirs[idx], row.iloc[0]), axis=1)
            if self.img_range is not None:
                img_labels = img_labels.iloc[self.img_range[0]:self.img_range[1]+1]
            # Apply skip: start after the first 'skip' number of rows
            if self.skip > 0:
                img_labels = img_labels.iloc[self.skip:]
            # Select specific rows based on the skip parameter
            img_labels = img_labels.iloc[::filter]
            # Limit the number of samples to max_samples if specified
            if max_samples is not None:
                img_labels = img_labels.iloc[:max_samples]
            # Determine if the images being fed are training or testing
            if test:
                self.img_labels = img_labels
            else:
                self.img_labels.append(img_labels)
        
        if isinstance(self.img_labels,list):
            # Concatenate all the DataFrames
            self.img_labels = pd.concat(self.img_labels, ignore_index=True)
        
    def __len__(self):
        return len(self.img_labels)
    
    def __getitem__(self, idx):
        img_path = self.img_labels.iloc[idx]['file_path']
        if not os.path.exists(img_path):
            raise FileNotFoundError(f"No file found for index {idx} at {img_path}.")
            
        image = read_image(img_path)
        label = self.img_labels.iloc[idx, 1]
        
        if self.transform:
            image = self.transform(image)
        if self.target_transform:
            label = self.target_transform(label)
            
        return image, label


import os
import csv

def create_csv_from_images(folder_path, csv_file_path):
    files = os.listdir(folder_path)
    png_files = sorted([f for f in files if f.endswith('.png')])

    with open(csv_file_path, 'w', newline='') as file:
        writer = csv.writer(file)
        writer.writerow(['Image_name', 'index'])
        
        for index, image_name in enumerate(png_files):
            writer.writerow([image_name, index])

# Name of the dataset to create .csv for
dataset_name = 'nordland-fall'

# Generate paths
folder_path = ''
csv_file_path = os.path.join('./VPRTempo/vprtempo/dataset', f'{dataset_name}.csv')

# Create .csv file
create_csv_from_images(folder_path, csv_file_path)


#MIT License

#Copyright (c) 2023 Adam Hines, Peter G Stratton, Michael Milford, Tobias Fischer

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'''
Imports
'''

import os
import torch

import numpy as np
import torch.nn as nn
import vprtempo.src.blitnet as bn
import torch.quantization as quantization
import torchvision.transforms as transforms

from tqdm import tqdm
from torch.utils.data import DataLoader
from torch.ao.quantization import QuantStub, DeQuantStub
from vprtempo.src.dataset import CustomImageDataset, ProcessImage

class VPRTempoQuantTrain(nn.Module):
    def __init__(self, args, dims, logger, num_modules, out_dim, out_dim_remainder=None):
        super(VPRTempoQuantTrain, self).__init__()

        # Set the arguments
        self.args = args
        for arg in vars(args):
            setattr(self, arg, getattr(args, arg))
        setattr(self, 'dims', dims)

        self.device = "cpu"
        self.logger = logger
        self.num_modules = num_modules
        self.quant = QuantStub()
        self.dequant = DeQuantStub()

        # Set the dataset file
        fields = self.database_dirs.split(',')
        if len(fields) > 1:
            self.dataset_file = []
            for field in fields:
                self.dataset_file.append(os.path.join('./vprtempo/dataset', f'{self.dataset}-{field}' + '.csv'))
        else:
            self.dataset_file = os.path.join('./vprtempo/dataset', f'{self.dataset}-{self.database_dirs}' + '.csv')

        # Layer dict to keep track of layer names and their order
        self.layer_dict = {}
        self.layer_counter = 0

        # Define layer architecture
        self.input = int(dims[0]*dims[1])
        self.feature = int(self.input * 2)
        if not out_dim_remainder is None:
            self.output = out_dim_remainder
        else:
            self.output = out_dim

        # Set the total timestep count
        self.database_dirs = [dir.strip() for dir in self.database_dirs.split(',')]
        self.location_repeat = len(self.database_dirs) # Number of times to repeat the locations
        if not out_dim_remainder is None:
            self.T = int(out_dim_remainder * self.location_repeat * self.epoch)
        else:
            self.T = int(self.max_module * self.location_repeat * self.epoch)

        """
        Define trainable layers here
        """
        self.add_layer(
            'feature_layer',
            dims=[self.input, self.feature],
            thr_range=[0, 0.5],
            fire_rate=[0.2, 0.9],
            ip_rate=0.15,
            stdp_rate=0.005,
            p=[0.1, 0.5],
            device=self.device
        )
        self.add_layer(
            'output_layer',
            dims=[self.feature, self.output],
            ip_rate=0.15,
            stdp_rate=0.005,
            spk_force=True,
            p=[1.0, 1.0],
            device=self.device
        )
        
    def add_layer(self, name, **kwargs):
        """
        Dynamically add a layer with given name and keyword arguments.
        
        :param name: Name of the layer to be added
        :type name: str
        :param kwargs: Hyperparameters for the layer
        """
        # Check for layer name duplicates
        if name in self.layer_dict:
            raise ValueError(f"Layer with name {name} already exists.")
        
        # Add a new SNNLayer with provided kwargs
        setattr(self, name, bn.SNNLayer(**kwargs))
        
        # Add layer name and index to the layer_dict
        self.layer_dict[name] = self.layer_counter
        self.layer_counter += 1                           

    def _anneal_learning_rate(self, layer, mod, itp, stdp):
        """
        Anneal the learning rate for the current layer.
        """
        if np.mod(mod, 100) == 0: # Modify learning rate every 100 timesteps
            pt = pow(float(self.T - mod) / self.T, 2)
            layer.eta_ip = torch.mul(itp, pt) # Anneal intrinsic threshold plasticity learning rate
            layer.eta_stdp = torch.mul(stdp, pt) # Anneal STDP learning rate
            
        return layer

    def train_model(self, train_loader, layer, model, model_num, prev_layers=None):
        """
        Train a layer of the network model.

        :param train_loader: Training data loader
        :param layer: Layer to train
        :param prev_layers: Previous layers to pass data through
        """

        # Initialize the tqdm progress bar
        pbar = tqdm(total=int(self.T),
                    desc="Training ",
                    position=0)
        
        # Initialize the learning rates for each layer (used for annealment)
        init_itp = layer.eta_ip.detach()
        init_stdp = layer.eta_stdp.detach()
        mod = 0  # Used to determine the learning rate annealment, resets at each epoch
        # idx scale factor for different modules
        idx_scale = (self.max_module*self.filter)*model_num
        # Run training for the specified number of epochs
        for _ in range(self.epoch):
            # Run training for the specified number of timesteps
            for spikes, labels in train_loader:
                spikes, labels = spikes.to(self.device), labels.to(self.device)
                idx = torch.round((labels - idx_scale) / self.filter) # Set output index for spike forcing
                # Pass through previous layers if they exist
                if prev_layers:
                    with torch.no_grad():
                        for prev_layer_name in prev_layers:
                            prev_layer = getattr(self, prev_layer_name) # Get the previous layer object
                            spikes = self.forward(spikes, prev_layer) # Pass spikes through the previous layer
                            spikes = bn.clamp_spikes(spikes, prev_layer) # Clamp spikes [0, 0.9]
                else:
                    prev_layer = None
                # Get the output spikes from the current layer
                pre_spike = spikes.detach() # Previous layer spikes for STDP
                spikes = self.forward(spikes, layer) # Current layer spikes
                spikes_noclp = spikes.detach() # Used for inhibitory homeostasis
                spikes = bn.clamp_spikes(spikes, layer) # Clamp spikes [0, 0.9]
                # Calculate STDP
                layer = bn.calc_stdp(pre_spike,spikes,spikes_noclp,layer, idx, prev_layer=prev_layer)
                # Adjust learning rates
                layer = self._anneal_learning_rate(layer, mod, init_itp, init_stdp)
                # Update the annealing mod & progress bar 
                mod += 1
                pbar.update(1)

        # Close the tqdm progress bar
        pbar.close()

    def forward(self, spikes, layer):
        """
        Compute the forward pass of the model.
    
        Parameters:
        - spikes (Tensor): Input spikes.
    
        Returns:
        - Tensor: Output after processing.
        """
        
        spikes = self.quant(spikes)
        spikes = layer.w(spikes)
        spikes = self.dequant(spikes)
        
        return spikes
    
    def save_model(self, models, model_out):    
        """
        Save the trained model to models output folder.
        """
        state_dicts = {}
        for i, model in enumerate(models):  # Assuming models_list is your list of models
            state_dicts[f'model_{i}'] = model.state_dict()

        torch.save(state_dicts, model_out)
            
def generate_model_name_quant(model):
    """
    Generate the model name based on its parameters.
    """
    return ("VPRTempoQuant" +
            str(model.input) +
            str(model.feature) +
            str(model.output) +
            str(model.num_modules) +
            '.pth')

def check_pretrained_model(model_name):
    """
    Check if a pre-trained model exists and prompt the user to retrain if desired.
    """
    if os.path.exists(os.path.join('./models', model_name)):
        prompt = "A network with these parameters exists, re-train network? (y/n):\n"
        retrain = input(prompt).strip().lower()
        return retrain == 'n'
    return False

def train_new_model_quant(models, model_name):
    """
    Train a new model.

    :param model: Model to train
    :param model_name: Name of the model to save after training
    """
    # Set first index model as the main model for parameters
    model = models[0]
    # Initialize the image transforms and datasets
    image_transform = transforms.Compose([
        ProcessImage(model.dims, model.patches)
    ])
    # Automatically generate user_input_ranges
    user_input_ranges = []
    start_idx = 0
    # Generate the image ranges for each module
    for _ in range(model.num_modules):
        range_temp = [start_idx, start_idx+((model.max_module-1)*model.filter)]
        user_input_ranges.append(range_temp)
        start_idx = range_temp[1] + model.filter

    # Keep track of trained layers to pass data through them
    trained_layers = [] 

    # Training each layer
    for layer_name, _ in sorted(model.layer_dict.items(), key=lambda item: item[1]):
        print(f"Training layer: {layer_name}")
        # Retrieve the layer object
        for i, model in enumerate(models):
            model.train()
            model.to(torch.device(model.device))
            layer = (getattr(model, layer_name))
            # Determine the maximum samples for the DataLoader
            if model.database_places < model.max_module:
                max_samples = model.database_places
            elif model.output < model.max_module:
                max_samples = model.output
            else:
                max_samples = model.max_module
            # Initialize new dataset with unique range for each module
            img_range=user_input_ranges[i]
            train_dataset = CustomImageDataset(annotations_file=model.dataset_file, 
                                    base_dir=model.data_dir,
                                    img_dirs=model.database_dirs,
                                    transform=image_transform,
                                    filter=models[0].filter,
                                    skip=models[0].skip,
                                    test=False,
                                    img_range=img_range,
                                    max_samples=max_samples)
            # Initialize the data loader
            train_loader = DataLoader(train_dataset, 
                                    batch_size=1, 
                                    shuffle=True,
                                    num_workers=8,
                                    persistent_workers=True)
            # Train the layers
            model.train_model(train_loader, layer, model, i, prev_layers=trained_layers)
        trained_layers.append(layer_name) 

    # Convert the model to evaluation mode
    for model in models:
        quantization.convert(model, inplace=True)

    # Save the model
    model.save_model(models,os.path.join('./vprtempo/models', model_name))      


import os
import torch
import logging

from datetime import datetime

def model_logger(): 
    """
    Configure the model logger
    """   
    now = datetime.now()
    output_base_folder = './vprtempo/output/'
    output_folder = output_base_folder + now.strftime("%d%m%y-%H-%M-%S")

    # Create the base output folder if it does not exist
    os.makedirs(output_base_folder, exist_ok=True)
    
    # Create the specific output folder
    os.mkdir(output_folder)
    # Create the logger
    logger = logging.getLogger("VPRTempo")
    if (logger.hasHandlers()):
        logger.handlers.clear()
    # Set the logger level
    logger.setLevel(logging.DEBUG)
    logging.basicConfig(filename=output_folder + "/logfile.log",
                        filemode="a+",
                        format="%(asctime)-15s %(levelname)-8s %(message)s")
    # Add the logger to the console (if specified)
    logger.addHandler(logging.StreamHandler())
        
    logger.info('')
    logger.info('██╗   ██╗██████╗ ██████╗ ████████╗███████╗███╗   ███╗██████╗  ██████╗') 
    logger.info('██║   ██║██╔══██╗██╔══██╗╚══██╔══╝██╔════╝████╗ ████║██╔══██╗██╔═══██╗')
    logger.info('██║   ██║██████╔╝██████╔╝   ██║   █████╗  ██╔████╔██║██████╔╝██║   ██║')
    logger.info('╚██╗ ██╔╝██╔═══╝ ██╔══██╗   ██║   ██╔══╝  ██║╚██╔╝██║██╔═══╝ ██║   ██║')
    logger.info(' ╚████╔╝ ██║     ██║  ██║   ██║   ███████╗██║ ╚═╝ ██║██║     ╚██████╔╝')
    logger.info('  ╚═══╝  ╚═╝     ╚═╝  ╚═╝   ╚═╝   ╚══════╝╚═╝     ╚═╝╚═╝      ╚═════╝ ')
    logger.info('-----------------------------------------------------------------------')
    logger.info('Temporally Encoded Spiking Neural Network for Visual Place Recognition v1.1.7')
    logger.info('Queensland University of Technology, Centre for Robotics')
    logger.info('')
    logger.info('© 2023 Adam D Hines, Peter G Stratton, Michael Milford, Tobias Fischer')
    logger.info('MIT license - https://github.com/QVPR/VPRTempo')
    logger.info('\\\\\\\\\\\\\\\\\\\\\\\\')
    logger.info('')
    if torch.cuda.is_available():
        logger.info('CUDA available: ' + str(torch.cuda.is_available()))
        current_device = torch.cuda.current_device()
        logger.info('Current device is: ' + str(torch.cuda.get_device_name(current_device)))
    else:
        logger.info('CUDA available: ' + str(torch.cuda.is_available()))
        logger.info('Current device is: CPU')
    logger.info('')

    return logger, output_folder

def model_logger_quant(): 
    """
    Configure the logger
    """   

    now = datetime.now()
    output_base_folder = './vprtempo/output/'
    output_folder = output_base_folder + now.strftime("%d%m%y-%H-%M-%S")

    # Create the base output folder if it does not exist
    os.makedirs(output_base_folder, exist_ok=True)
    
    # Create the specific output folder
    os.mkdir(output_folder)
    # Create the logger
    logger = logging.getLogger("VPRTempo")
    if (logger.hasHandlers()):
        logger.handlers.clear()
    # Set the logger level
    logger.setLevel(logging.DEBUG)
    logging.basicConfig(filename=output_folder + "/logfile.log",
                        filemode="a+",
                        format="%(asctime)-15s %(levelname)-8s %(message)s")
    # Add the logger to the console (if specified)
    logger.addHandler(logging.StreamHandler())
        
    logger.info('')

    logger.info('██╗   ██╗██████╗ ██████╗ ████████╗███████╗███╗   ███╗██████╗  ██████╗        ██████╗ ██╗   ██╗ █████╗ ███╗   ██╗████████╗')
    logger.info('██║   ██║██╔══██╗██╔══██╗╚══██╔══╝██╔════╝████╗ ████║██╔══██╗██╔═══██╗      ██╔═══██╗██║   ██║██╔══██╗████╗  ██║╚══██╔══╝')
    logger.info('██║   ██║██████╔╝██████╔╝   ██║   █████╗  ██╔████╔██║██████╔╝██║   ██║█████╗██║   ██║██║   ██║███████║██╔██╗ ██║   ██║')   
    logger.info('╚██╗ ██╔╝██╔═══╝ ██╔══██╗   ██║   ██╔══╝  ██║╚██╔╝██║██╔═══╝ ██║   ██║╚════╝██║▄▄ ██║██║   ██║██╔══██║██║╚██╗██║   ██║')   
    logger.info(' ╚████╔╝ ██║     ██║  ██║   ██║   ███████╗██║ ╚═╝ ██║██║     ╚██████╔╝      ╚██████╔╝╚██████╔╝██║  ██║██║ ╚████║   ██║')   
    logger.info('  ╚═══╝  ╚═╝     ╚═╝  ╚═╝   ╚═╝   ╚══════╝╚═╝     ╚═╝╚═╝      ╚═════╝        ╚══▀▀═╝  ╚═════╝ ╚═╝  ╚═╝╚═╝  ╚═══╝   ╚═╝')                                                                                                                     
    logger.info('-----------------------------------------------------------------------')
    logger.info('Temporally Encoded Spiking Neural Network for Visual Place Recognition v1.1.7')
    logger.info('Queensland University of Technology, Centre for Robotics')
    logger.info('')
    logger.info('© 2023 Adam D Hines, Peter G Stratton, Michael Milford, Tobias Fischer')
    logger.info('MIT license - https://github.com/QVPR/VPRTempo')
    logger.info('\\\\\\\\\\\\\\\\\\\\\\\\')
    logger.info('')
    logger.info('Quantization enabled')
    logger.info('Current device is: CPU')
    logger.info('')

    return logger, output_folder



'''
Imports
'''
import os
import re
import shutil
import zipfile
import sys
sys.path.append('..//dataset')

from os import walk

def nord_sort():
    # load and sort the file names in order, not by how OS indexes them
    def atoi(text):
        return int(text) if text.isdigit() else text
    
    def natural_keys(text):
        return [ atoi(c) for c in re.split(r'(\d+)', text) ]
    
    # set the base path to the location of the downloaded Nordland datasets
    basePath = '../dataset/'
    assert(os.path.isdir(basePath)),"Please set the basePath to the location of the downloaded Nordland datasets"
    
    # define the subfolders of the Nordland datasets
    subPath = ["spring_images_train/section1/","spring_images_train/section2/",
               "fall_images_train/section1/","fall_images_train/section2/",
               "winter_images_train/section1/","winter_images_train/section2/",
               "summer_images_train/section1/","summer_images_train/section2/"]
    
    # set the desired output folder for unzipping and organization
    outDir = '../dataset/'
    assert(os.path.isdir(outDir)),"Please set the outDir to the desired output location for unzipping the Nordland datasets"
    
    # define output paths for the data
    outPath = [os.path.join(outDir,"spring/"),os.path.join(outDir,"fall/"),
               os.path.join(outDir,"winter/"),os.path.join(outDir,"summer/")]
    
    # check for existence of the zip folders, throw exception if missing
    zipNames = ["spring_images_train.zip","fall_images_train.zip",
                "winter_images_train.zip","summer_images_train.zip"]
    for n in zipNames:
        if not os.path.exists(basePath+n):
            raise Exception('Please ensure dataset .zip folders have been downloaded') 
    
    # check if nordland data folders have already been unzipped
    zip_flag = []
    for n, ndx in enumerate(range(0,len(subPath),2)):
        print('Unzipping '+zipNames[n])
        if os.path.exists(basePath+subPath[ndx]):
            # check if the folder contains any files
            file_lst = os.listdir(basePath+subPath[ndx])
            # remove folder if it is empty and unzip the data folder
            if len(file_lst) == 0:
                shutil.rmtree(basePath+subPath[ndx].replace('section1/',''))
                with zipfile.ZipFile(basePath+zipNames[n],"r") as zip_ref:
                    zip_ref.extractall(basePath)
        else:
            with zipfile.ZipFile(basePath+zipNames[n],"r") as zip_ref:
                zip_ref.extractall(basePath)
                
    # load image paths
    tempPaths = []
    imgPaths = []
    for n in range(0,len(subPath)):
        tempPaths = []
        for (path, dir_names, file_names) in walk(basePath+subPath[n]):
            tempPaths.extend(file_names)   
        # sort image names
        tempPaths.sort(key=natural_keys)
        tempPaths = [basePath+subPath[n]+s for s in tempPaths]
        imgPaths = imgPaths + tempPaths
    
    # if output folders already exist, delete them
    for n in outPath:
        if os.path.exists(n):
            shutil.rmtree(n)
            print('Removed pre-existing output folder')
        
    # rename and move the training data to match the nordland_imageNames.txt file
    for n in outPath:
        os.mkdir(n)
    for n, filename in enumerate(imgPaths):
        base = os.path.basename(filename)
        split_base = os.path.splitext(base)
        if int(split_base[0]) < 10:
            my_dest = "images-0000"+split_base[0] + ".png"
        elif int(split_base[0]) < 100:
            my_dest = "images-000"+split_base[0] + ".png"
        elif int(split_base[0]) < 1000:
            my_dest = "images-00"+split_base[0] + ".png"
        elif int(split_base[0]) < 10000:
            my_dest = "images-0"+split_base[0] + ".png"
        else:
            my_dest = "images-"+split_base[0] + ".png"
        if "spring" in filename:
            out = outPath[0]
        elif "fall" in filename:
            out = outPath[1]
        elif "winter" in filename:
            out = outPath[2]
        else:
            out = outPath[-1]
    
        fileDest = out + my_dest   
        os.rename(filename, fileDest)
    
    # remove the empty folders
    for n, ndx in enumerate(subPath):
        if n%2 == 0:
            shutil.rmtree(basePath+ndx.replace('section1/',''))
        else:
            continue
    
    print('Finished unzipping and organizing Nordland dataset')


#MIT License

#Copyright (c) 2023 Adam Hines, Peter Stratton, Michael Milford, Tobias Fischer

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'''
Imports
'''
import torch

import torch.nn as nn
import numpy as np


class SNNLayer(nn.Module):
    def __init__(self, dims=[0,0],thr_range=[0,0],fire_rate=[0,0],ip_rate=0,
                 stdp_rate=0,const_inp=[0,0],p=[1,1],spk_force=False,device=None,inference=False,args=None):
        super(SNNLayer, self).__init__()
        """
        dims: [input, output] dimensions of the layer
        thr_range: [min, max] range of thresholds
        fire_rate: [min, max] range of firing rates
        ip_rate: learning rate for input threshold plasticity
        stdp_rate: learning rate for stdp
        const_inp: [min, max] range of constant input
        p: [min, max] range of connection probabilities
        spk_force: boolean to force spikes
        """

        # Device
        self.device = device
        # Add different parameters depending if trainnig or running inference model
        if inference: # If running inference model
            self.w = nn.Linear(dims[0], dims[1], bias=False) # Combined weight tensors
            self.w.to(device)
            self.thr = nn.Parameter(torch.zeros([1, dims[-1]], 
                                                device=self.device).uniform_(thr_range[0], 
                                                                            thr_range[1]))
        else: # If training new model
            # Check constraints etc
            if np.isscalar(thr_range): thr_range = [thr_range, thr_range]
            if np.isscalar(fire_rate): fire_rate = [fire_rate, fire_rate]
            if np.isscalar(const_inp): const_inp = [const_inp, const_inp]
            
            # Initialize Tensors
            self.x = torch.zeros([1, dims[-1]], device=self.device)
            self.eta_ip = torch.tensor(ip_rate, device=self.device)
            self.eta_stdp = torch.tensor(stdp_rate, device=self.device)
            
            # Initialize Parameters
            self.thr = nn.Parameter(torch.zeros([1, dims[-1]], 
                                                device=self.device).uniform_(thr_range[0], 
                                                                            thr_range[1]))
            self.fire_rate = torch.zeros([1,dims[-1]], device=self.device).uniform_(fire_rate[0], fire_rate[1])
            
            # Sequentially set the feature firing rates (if any)
            if not torch.all(self.fire_rate==0).item():
                fstep = (fire_rate[1]-fire_rate[0])/dims[-1]
                
                for i in range(dims[-1]):
                    self.fire_rate[:,i] = fire_rate[0]+fstep*(i+1)
                    
            self.have_rate = torch.any(self.fire_rate[:,0] > 0.0).to(self.device)
            self.const_inp = torch.zeros([1, dims[-1]], device=self.device).uniform_(const_inp[0], const_inp[1])
            self.p = p
            self.dims = dims
            
            # Additional State Variables
            self.set_spks = []
            self.sspk_idx = 0
            self.spikes = torch.empty([], dtype=torch.float64)
            self.spk_force = spk_force
        
            # Create the excitatory weights
            self.exc = nn.Linear(dims[0], dims[1], bias=False)
            self.exc.weight = self.addWeights(dims=dims,
                                                W_range=[0,1], 
                                                p=p[0],
                                                device=device)
            
            # Create the inhibitory weights
            self.inh = nn.Linear(dims[0], dims[1], bias=False)
            self.inh.weight = self.addWeights(dims=dims,
                                                W_range=[-1,0], 
                                                p=p[-1],
                                                device=device)
            
            # Output boolean reference of which neurons have connection weights
            self.havconnExc = self.exc.weight > 0
            self.havconnInh = self.inh.weight < 0

            # Combine weights into a single tensor
            self.w = nn.Linear(dims[0], dims[1], bias=False)
            self.w.weight = nn.Parameter(torch.add(self.exc.weight, self.inh.weight))

            self.havconnCombinedExc = self.w.weight > 0
            self.havconnCombinedInh = self.w.weight < 0

            del self.exc, self.inh

    def addWeights(self,W_range=[0,0],p=[0,0],dims=[0,0],device=None):

        # Get torch device  
        device = device
    
        # Check constraints etc
        if np.isscalar(W_range): W_range = [W_range,W_range]
        
        # Determine dimensions of the weight matrices
        nrow = dims[1]
        ncol = dims[0]
        
        # Calculate mean and std for normal distributions
        Wmn = (W_range[0]+W_range[1])/2.0
        Wsd = (W_range[1]-W_range[0])/6.0
        
        # Initialize weights as empty tensors
        W = torch.empty((0, nrow, ncol), device=device)
        
        # Normally disribute random weights
        W = torch.empty(nrow, ncol, device=device).normal_(mean=Wmn, std=Wsd)
            
        # Remove inappropriate weights based on sign from W_range
        if W_range[-1] != 0:
            # For excitatory weights
            W[W < 0] = 0.0
        else:
            # For inhibitory weights
            W[W > 0] = 0.0
    
        # Remove weights based on connection probability
        setzero = np.random.rand(nrow,ncol) > p
        if setzero.any():
           W[setzero] = 0.0
    
        # Normalise the weights
        nrm = torch.linalg.norm(W[len(W)-1],ord=1,axis=0)
        nrm[nrm==0.0] = 1.0
        W = nn.Parameter(W/nrm)
    
        return W
        
def add_input(spikes, layer):
    
    # Add the constant input
    spikes += layer.const_inp
    
    return spikes

def clamp_spikes(spikes, layer):
    # Clamp outputs between 0 and 0.9 after subtracting thresholds from input
    spikes = torch.clamp(torch.sub(spikes, layer.thr), min=0.0, max=0.9)
        
    return spikes

def calc_stdp(prespike, spikes, noclp, layer, idx, prev_layer=None):
    # Spike Forcing has special rules to make calculated and forced spikes match
    if layer.spk_force:
        
        # Get layer dimensions
        shape = layer.w.weight.data.shape
        
        # Get the output neuron index
        idx_sel = torch.arange(int(idx[0]), int(idx[0]) + 1, 
                               device=layer.device, 
                               dtype=int)

        # Difference between forced and calculated spikes
        layer.x = torch.full_like(layer.x, 0)
        xdiff = layer.x.index_fill_(-1, idx_sel, 0.5) - spikes
        xdiff.clamp(min=0.0, max=0.9)

        # Pre and Post spikes tiled across and down for all synapses
        if prev_layer.fire_rate == None:
            mpre = prespike
        else:
            # Modulate learning rate by firing rate (low firing rate = high learning rate)
            mpre = prespike/prev_layer.fire_rate
            
        # Tile out pre- and post- spikes for STDP weight updates    
        pre = torch.tile(torch.reshape(mpre, (shape[1], 1)), (1, shape[0]))
        post = torch.tile(xdiff, (shape[1], 1))

        # Apply the weight changes
        layer.w.weight.data += ((pre * post * layer.havconnCombinedExc.T) * 
                                       layer.eta_stdp).T
        layer.w.weight.data += ((-pre * post * layer.havconnCombinedInh.T) * 
                                       (layer.eta_stdp * -1)).T

    # Normal STDP
    else:
        
        # Get layer dimensions
        shape = layer.w.weight.data.shape
        
        # Tile out pre- and post-spikes
        pre = torch.tile(torch.reshape(prespike, (shape[1], 1)), (1, shape[0]))
        post = torch.tile(spikes, (shape[1], 1))
        
        # Apply positive and negative weight changes
        layer.w.weight.data += (((0.5 - post) * (pre > 0) * (post > 0) * 
                                  layer.havconnCombinedExc.T) * layer.eta_stdp).T
        layer.w.weight.data += (((0.5 - post) * (pre > 0) * 
                                  (post > 0) * layer.havconnCombinedInh.T) * (layer.eta_stdp * -1)).T
    
    # Remove negative weights for excW and positive for inhW
    layer.w.weight.data[layer.havconnCombinedExc] = layer.w.weight.data[layer.havconnCombinedExc].clamp(min=1e-06, max=10)
    layer.w.weight.data[layer.havconnCombinedInh] = layer.w.weight.data[layer.havconnCombinedInh].clamp(min=-10, max=-1e-06)

    # Check if layer has target firing rate and an ITP learning rate
    if layer.have_rate and layer.eta_ip > 0.0:
        
        # Replace the original layer.thr with the updated one
        layer.thr.data += layer.eta_ip * (layer.x - layer.fire_rate)
        layer.thr.data[layer.thr.data < 0] = 0
    
    # Check if layer has inhibitory weights and an stdp learning rate
    if torch.any(layer.w.weight.data).item() and layer.eta_stdp != 0:
        
        # Normalize the inhibitory weights using homeostasis
        inhW = layer.w.weight.data.T.clone()
        inhW[inhW>0] = 0
        layer.w.weight.data += (torch.mul(noclp,inhW) * layer.eta_stdp*50).T
        #layer.w.weight.data[layer.w.weight.data > 0.0] = -1e-06

    return layer


#MIT License

#Copyright (c) 2023 Adam Hines, Peter G Stratton, Michael Milford, Tobias Fischer

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

'''
Imports
'''

import os
import json
import torch
import random

import numpy as np
import torch.nn as nn
import matplotlib.pyplot as plt
import vprtempo.src.blitnet as bn

from tqdm import tqdm
from prettytable import PrettyTable
from torch.utils.data import DataLoader, Subset
from vprtempo.src.metrics import recallAtK, createPR
from torch.ao.quantization import QuantStub, DeQuantStub
from vprtempo.src.dataset import CustomImageDataset, ProcessImage

#from main import parse_network

class VPRTempoQuant(nn.Module):
    def __init__(self, args, dims, logger, num_modules, output_folder, out_dim, out_dim_remainder=None):
        super(VPRTempoQuant, self).__init__()

        # Set the args
        if args is not None:
            self.args = args
            for arg in vars(args):
                setattr(self, arg, getattr(args, arg))
        setattr(self, 'dims', dims)

        # Set the device
        self.device = "cpu"

        # Set input args
        self.logger = logger
        self.num_modules = num_modules
        self.output_folder = output_folder

        self.quant = QuantStub()
        self.dequant = DeQuantStub()

        # Set the dataset file
        self.dataset_file = os.path.join('./vprtempo/dataset', f'{self.dataset}-{self.query_dir}' + '.csv')  
        self.query_dir = [dir.strip() for dir in self.query_dir.split(',')]

        # Layer dict to keep track of layer names and their order
        self.layer_dict = {}
        self.layer_counter = 0
        self.database_dirs = [dir.strip() for dir in self.database_dirs.split(',')]

        # Define layer architecture
        self.input = int(self.dims[0]*self.dims[1])
        self.feature = int(self.input * 2)

        # Output dimension changes for final module if not an even distribution of places
        if not out_dim_remainder is None:
            self.output = out_dim_remainder
        else:
            self.output = out_dim

        """
        Define trainable layers here
        """
        self.add_layer(
            'feature_layer',
            dims=[self.input, self.feature],
            device=self.device,
            inference=True
        )
        self.add_layer(
            'output_layer',
            dims=[self.feature, self.output],
            device=self.device,
            inference=True
        )
        
    def add_layer(self, name, **kwargs):
        """
        Dynamically add a layer with given name and keyword arguments.
        
        :param name: Name of the layer to be added
        :type name: str
        :param kwargs: Hyperparameters for the layer
        """
        # Check for layer name duplicates
        if name in self.layer_dict:
            raise ValueError(f"Layer with name {name} already exists.")
        
        # Add a new SNNLayer with provided kwargs
        setattr(self, name, bn.SNNLayer(**kwargs))
        
        # Add layer name and index to the layer_dict
        self.layer_dict[name] = self.layer_counter
        self.layer_counter += 1                           

    def evaluate(self, models, test_loader, layers=None):
        """
        Run the inferencing model and calculate the accuracy.

        :param test_loader: Testing data loader
        :param layers: Layers to pass data through
        """
        # Determine the Hardtahn max value
        maxSpike = (1//models[0].quant.scale).item()
        # Define the sequential inference model
        self.inferences = []
        for model in models:
            self.inferences.append(nn.Sequential(
                model.feature_layer.w,
                model.output_layer.w,
            ))
        # Initialize the tqdm progress bar
        pbar = tqdm(total=self.query_places,
                    desc="Running the test network",
                    position=0)
        # Initiliaze the output spikes variable
        out = []
        labels = []
        # Run inference for the specified number of timesteps
        for spikes, label in test_loader:
            # Set device
            spikes = spikes.to(self.device)
            labels.append(label.detach().item())
            # Pass through previous layers if they exist
            spikes = self.forward(spikes)
            # Add output spikes to list
            out.append(spikes.detach().cpu().tolist())
            pbar.update(1)

        # Close the tqdm progress bar
        pbar.close()

        # Rehsape output spikes into a similarity matrix
        out = np.reshape(np.array(out),(model.query_places,model.database_places))

        # Create GT matrix
        GT = np.zeros((model.query_places,model.database_places), dtype=int)
        for n, ndx in enumerate(labels):
            if model.skip != 0 and not model.query_places < model.database_places:
                ndx = ndx - model.skip
            if model.filter !=1:
                ndx = ndx//model.filter
            GT[n,ndx] = 1

        # Create GT soft matrix
        if model.GT_tolerance > 0:
            GTsoft = np.zeros((model.query_places,model.database_places), dtype=int)
            for n, ndx in enumerate(labels):
                if model.skip != 0 and not model.query_places < model.database_places:
                    ndx = ndx - model.skip
                if model.filter !=1:
                    ndx = ndx//model.filter
                GTsoft[n, ndx] = 1
                # Apply tolerance
                for i in range(max(0, n - model.GT_tolerance), min(model.query_places, n + model.GT_tolerance + 1)):
                    GTsoft[i, ndx] = 1
        else:
            GTsoft = None
        
        # If user specified, generate a PR curve
        if model.PR_curve:
            # Create PR curve
            P, R = createPR(out, GThard=GT, GTsoft=GTsoft, matching='single', n_thresh=100)
            # Combine P and R into a list of lists
            PR_data = {
                    "Precision": P,
                    "Recall": R
                }
            output_file = "PR_curve_data.json"
            # Construct the full path
            full_path = f"{model.output_folder}/{output_file}"
            # Write the data to a JSON file
            with open(full_path, 'w') as file:
                json.dump(PR_data, file) 
            # Plot PR curve
            plt.plot(R,P)    
            plt.xlabel('Recall')
            plt.ylabel('Precision')
            plt.title('Precision-Recall Curve')
            plt.show()

        if model.sim_mat:
            # Create a figure and a set of subplots
            fig, axs = plt.subplots(1, 3, figsize=(15, 5))

            # Plot each matrix using matshow
            cax1 = axs[0].matshow(out, cmap='viridis')
            fig.colorbar(cax1, ax=axs[0], shrink=0.8)
            axs[0].set_title('Similarity matrix')

            cax2 = axs[1].matshow(GT, cmap='plasma')
            fig.colorbar(cax2, ax=axs[1], shrink=0.8)
            axs[1].set_title('GT')

            cax3 = axs[2].matshow(GTsoft, cmap='inferno')
            fig.colorbar(cax3, ax=axs[2], shrink=0.8)
            axs[2].set_title('GT-soft')

            # Adjust layout
            plt.tight_layout()
            plt.show()

        # Recall@N
        N = [1,5,10,15,20,25] # N values to calculate
        R = [] # Recall@N values
        # Calculate Recall@N
        for n in N:
            R.append(round(recallAtK(out,GThard=GT,GTsoft=GTsoft,K=n),2))
        # Print the results
        table = PrettyTable()
        table.field_names = ["N", "1", "5", "10", "15", "20", "25"]
        table.add_row(["Recall", R[0], R[1], R[2], R[3], R[4], R[5]])
        self.logger.info(table)

    def forward(self, spikes):
        """
        Compute the forward pass of the model.
    
        Parameters:
        - spikes (Tensor): Input spikes.
    
        Returns:
        - Tensor: Output after processing.
        """
        spikes = self.quant(spikes)
        in_spikes = spikes.detach().clone()
        outputs = []  # List to collect output tensors

        for inference in self.inferences:
            out_spikes = inference(in_spikes)
            outputs.append(out_spikes)  # Append the output tensor to the list

        # Concatenate along the desired dimension
        concatenated_output = torch.cat(outputs, dim=1)
        spikes = self.dequant(concatenated_output)

        return spikes
        
    def load_model(self, models, model_path):
        """
        Load pre-trained model and set the state dictionary keys.
        """
        combined_state_dict = torch.load(model_path, map_location=self.device)

        for i, model in enumerate(models):  # models_classes is a list of model classes

            model.load_state_dict(combined_state_dict[f'model_{i}'])
            model.eval()  # Set the model to inference mode

def run_inference_quant(models, model_name):
    """
    Run inference on a pre-trained model.

    :param models: Models to run inference on, each model is a VPRTempo module
    :param model_name: Name of the model to load
    """
    # Set first index model as the main model for parameters
    model = models[0]
    # Initialize the image transforms
    image_transform = ProcessImage(model.dims, model.patches)
    
    # Initialize the test dataset
    test_dataset = CustomImageDataset(annotations_file=model.dataset_file, 
                                      base_dir=model.data_dir,
                                      img_dirs=model.query_dir,
                                      transform=image_transform,
                                      max_samples=model.query_places,
                                      filter=model.filter,
                                      skip=model.skip)
    
    # Initialize the data loader
    test_loader = DataLoader(test_dataset, 
                             batch_size=1,
                             num_workers=8,
                             persistent_workers=True)

    # Load the model
    model.load_model(models, os.path.join('./vprtempo/models', model_name))

    # Use evaluate method for inference accuracy
    with torch.no_grad():
        model.evaluate(models, test_loader)


import os, sys

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
# Get the long description from the README file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# define the base requires needed for the repo
requirements = [ 
                'torch',
                'torchvision',
                'numpy',
                'pandas',
                'tqdm',
                'prettytable',
                'matplotlib'
                ]

# define the setup
setup(
    name="VPRTempo",
    version="1.1.7",
    description='VPRTempo: A Fast Temporally Encoded Spiking Neural Network for Visual Place Recognition',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Adam D Hines, Peter G Stratton, Michael Milford and Tobias Fischer',
    author_email='adam.hines@qut.edu.au',
    url='https://github.com/QVPR/VPRTempo',
    license='MIT',
    install_requires=requirements,
    python_requires='>=3.6',
    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
      ],
    packages=find_packages(),
    keywords=['python', 'place recognition', 'spiking neural networks',
              'computer vision', 'robotics'],
    scripts=['main.py'],
)
